#include "postgres.h"
#include "fmgr.h"
#include "miscadmin.h"
#include <math.h>

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif

Datum byte_entropy(PG_FUNCTION_ARGS);
Datum bit_entropy(PG_FUNCTION_ARGS);
Datum chi_square(PG_FUNCTION_ARGS);

#define BYTE_SIZE 256
#define BIT_SIZE 2

/* This function calculates the byte entropy of a given bytea:
 *   http://en.wikipedia.org/wiki/Entropy_%28information_theory%29#Definition
 * return value in [0.0, 8.0]
 */
PG_FUNCTION_INFO_V1(byte_entropy);
Datum
byte_entropy(PG_FUNCTION_ARGS) {
    bytea *data = PG_GETARG_BYTEA_P(0);
    unsigned char *ptr = (unsigned char *) VARDATA(data);
    int32 tcount = VARSIZE(data) - VARHDRSZ, i;
    float8 entropy = 0, p;
    unsigned char c;
   
    // initialize character counts
    int ccount[BYTE_SIZE];
    for (i = 0; i < BYTE_SIZE; i++) {
        ccount[i] = 0;
    }

    // count characters
    for (i = VARSIZE(data) - VARHDRSZ; i; i--) {
        if (!i%1000) CHECK_FOR_INTERRUPTS();
        c = *ptr++;
        ccount[c]++;
    }

    if (tcount != 0) { // entropy is 0 if bytea was empty
        // calculate the probabilities for each char and sum entropy
        for (i = 0; i < BYTE_SIZE; i++) {
            p = ccount[i] / (float8) tcount;
            if (p == 0)
                continue;   // if p = 0 then log(p) = 0

            entropy -= p * log(p) / log(2);
        }
    }

    PG_RETURN_FLOAT8(entropy);
}

/* This function calculates the bit entropy of a given bytea:
 *   http://en.wikipedia.org/wiki/Entropy_%28information_theory%29#Definition
 * return value in [0.0, 1.0]
 */
PG_FUNCTION_INFO_V1(bit_entropy);
Datum
bit_entropy(PG_FUNCTION_ARGS) {
    bytea *data = PG_GETARG_BYTEA_P(0);
    unsigned char *ptr = (unsigned char *) VARDATA(data);
    int32 tcount = (VARSIZE(data) - VARHDRSZ) * 8, i, j;
    float8 entropy = 0, p;
    unsigned char c;
   
    // initialize character counts
    int ccount[BIT_SIZE];
    for (i = 0; i < BIT_SIZE; i++) {
        ccount[i] = 0;
    }

    // count characters
    for (i = VARSIZE(data) - VARHDRSZ; i; i--) {
        if (!i%1000) CHECK_FOR_INTERRUPTS();
        c = *ptr++;
        for (j = 0; j < 8; j++) {
            if (c & (1 << j)) {
                ccount[1]++;
            } else {
                ccount[0]++;
            }
        }
    }

    if (tcount != 0) { // entropy is 0 if bytea was empty
        // calculate the probabilities for each bit and sum entropy
        for (i = 0; i < BIT_SIZE; i++) {
            p = ccount[i] / (float8) tcount;
            if (p == 0)
                continue;   // if p = 0 then log(p) = 0

            entropy -= p * log(p) / log(2);
        }
    }

    PG_RETURN_FLOAT8(entropy);
}

PG_FUNCTION_INFO_V1(chi_square);
Datum
chi_square(PG_FUNCTION_ARGS) {
    bytea *data = PG_GETARG_BYTEA_P(0);
    unsigned char *ptr = (unsigned char *) VARDATA(data);
    int32 tcount = VARSIZE(data) - VARHDRSZ, i;
    float8 chisq = 0, cexp;
    unsigned char c;
   
    // initialize character counts
    int ccount[BYTE_SIZE];
    for (i = 0; i < BYTE_SIZE; i++) {
        ccount[i] = 0;
    }

    // count characters
    for (i = VARSIZE(data) - VARHDRSZ; i; i--) {
        if (!i%1000) CHECK_FOR_INTERRUPTS();
        c = *ptr++;
        ccount[c]++;
    }

    if (tcount != 0) { // chisq is 0 if bytea was empty
        // calculate the chi-square per char
        for (i = 0; i < BYTE_SIZE; i++) {
            cexp = tcount / (float8) BYTE_SIZE;
            chisq += (ccount[i] - cexp) * (ccount[i] - cexp) / cexp;
        }
    }

    PG_RETURN_FLOAT8(chisq);
}
